from typing import Mapping, Tuple, Union, List, Dict, Any

from tornado import gen
from tornado import httpclient

from . import abc as gh_abc


class GitHubAPI(gh_abc.GitHubAPI):
    async def _request(
        self, method: str, url: str, headers: Mapping[str, str], body: bytes = b""
    ) -> Tuple[int, Mapping[str, str], bytes]:
        """Make an HTTP request."""
        # Setting 'body' to None fails type checking, so only add a 'body' argument if necessary.
        args: List[Union[str, Dict[Any, Any], bytes]] = [url, method, dict(headers)]
        if method != "GET" and body:
            args.append(body)
        # The below line is skipped from mypy because Tornado's HTTPRequest signature
        # requires many types of arguments some of which are internal to it
        # adding all of them to the `args` would be impractical.
        request = httpclient.HTTPRequest(*args)  # type: ignore
        # Since Tornado has designed AsyncHTTPClient to be a singleton, there's
        # no reason not to simply instantiate it every time.
        client = httpclient.AsyncHTTPClient()
        response = await client.fetch(request, raise_error=False)
        return response.code, response.headers, response.body

    async def sleep(self, seconds: float) -> None:
        """Sleep for the specified number of seconds."""
        await gen.sleep(seconds)



import datetime
import http
import json
import pathlib


import pytest

from gidgethub import (
    BadRequest,
    BadRequestUnknownError,
    GitHubBroken,
    HTTPException,
    InvalidField,
    RateLimitExceeded,
    RedirectionException,
    ValidationError,
    ValidationFailure,
)
from gidgethub import sansio


class TestValidateEvent:
    """Tests for gidgethub.sansio.validate_event()."""

    secret = "123456"
    payload = b"gidget"
    hash_signature = "091319196718d5bcb1c20ad25fc890597423ecdbad1f947f560afd643b5000de"
    signature = "sha256=" + hash_signature

    def test_malformed_signature(self):
        """Error out if the signature doesn't start with "sha256=" or "sha1="."""
        with pytest.raises(ValidationFailure):
            sansio.validate_event(
                self.payload, secret=self.secret, signature=self.hash_signature
            )

    def test_validation(self):
        """Success case."""
        sansio.validate_event(
            self.payload, secret=self.secret, signature=self.signature
        )

    def test_failure(self):
        with pytest.raises(ValidationFailure):
            sansio.validate_event(
                self.payload + b"!", secret=self.secret, signature=self.signature
            )


class TestEvent:
    """Tests for gidgethub.sansio.Event."""

    data = {"action": "opened"}
    data_bytes = b'{"action": "opened"}'
    secret = "123456"
    headers = {
        "content-type": "application/json",
        "x-github-event": "pull_request",
        "x-github-delivery": "72d3162e-cc78-11e3-81ab-4c9367dc0958",
        "x-hub-signature": "sha1=c28e33b2e56e548956c446e890929a6cbec3ac89",
        "x-hub-signature-256": "sha256=0340d06469a1b35662ebd7a67ea2c0c328239f319f1cfafb221451de629e0430",
    }

    def check_event(self, event):
        """Check that an event matches the test data provided by the class."""
        assert event.event == self.headers["x-github-event"]
        assert event.delivery_id == self.headers["x-github-delivery"]
        assert event.data == self.data

    def test_init(self):
        ins = sansio.Event(
            self.data,
            event=self.headers["x-github-event"],
            delivery_id=self.headers["x-github-delivery"],
        )
        self.check_event(ins)

    def test_from_http_json(self):
        """Construct an event from complete HTTP information."""
        event = sansio.Event.from_http(
            self.headers, self.data_bytes, secret=self.secret
        )
        self.check_event(event)

    def test_from_http_urlencoded(self):
        headers, body = sample("ping_urlencoded", 200)
        event = sansio.Event.from_http(headers, body)
        assert event.data["zen"] == "Keep it logically awesome."

    def test_from_http_no_content_type(self):
        """Only accept data when content-type is application/json."""
        headers_no_content_type = self.headers.copy()
        del headers_no_content_type["content-type"]
        with pytest.raises(BadRequest):
            sansio.Event.from_http(
                headers_no_content_type, self.data_bytes, secret=self.secret
            )

    def test_from_http_unknown_content_type(self):
        headers = headers = {
            "content-type": "image/png",
            "x-github-event": "pull_request",
            "x-github-delivery": "72d3162e-cc78-11e3-81ab-4c9367dc0958",
        }
        with pytest.raises(BadRequest):
            sansio.Event.from_http(headers, self.data_bytes)
        pass

    def test_from_http_missing_secret(self):
        """Signature but no secret raises ValidationFailure."""
        with pytest.raises(ValidationFailure):
            sansio.Event.from_http(self.headers, self.data_bytes)

    def test_from_http_missing_signature(self):
        """Secret but no signature raises ValidationFailure."""
        headers_no_sig = self.headers.copy()
        del headers_no_sig["x-hub-signature-256"]
        del headers_no_sig["x-hub-signature"]
        with pytest.raises(ValidationFailure):
            sansio.Event.from_http(headers_no_sig, self.data_bytes, secret=self.secret)

    def test_from_http_bad_signature(self):
        with pytest.raises(ValidationFailure):
            sansio.Event.from_http(
                self.headers, self.data_bytes, secret=self.secret + "no secret"
            )

    def test_from_http_sha1_signature(self):
        headers = self.headers.copy()
        del headers["x-hub-signature-256"]
        event = sansio.Event.from_http(headers, self.data_bytes, secret=self.secret)
        self.check_event(event)

    def test_from_http_no_signature(self):
        headers = self.headers.copy()
        del headers["x-hub-signature-256"]
        del headers["x-hub-signature"]
        event = sansio.Event.from_http(headers, self.data_bytes)
        self.check_event(event)


class TestAcceptFormat:
    """Tests for gidgethub.sansio.accept_format()."""

    def test_defaults(self):
        assert sansio.accept_format() == "application/vnd.github.v3+json"

    def test_format(self):
        expect = "application/vnd.github.v3.raw+json"
        assert sansio.accept_format(media="raw") == expect

    def test_no_json(self):
        expect = "application/vnd.github.v3.raw"
        assert sansio.accept_format(media="raw", json=False) == expect

    def test_version(self):
        expect = "application/vnd.github.cloak-preview+json"
        assert sansio.accept_format(version="cloak-preview") == expect


class TestCreateHeaders:
    """Tests for gidgethub.sansio.create_headers()."""

    def test_common_case(self):
        user_agent = "brettcannon"
        oauth_token = "secret"
        headers = sansio.create_headers(user_agent, oauth_token=oauth_token)
        assert headers["user-agent"] == user_agent
        assert headers["authorization"] == f"token {oauth_token}"

    def test_api_change(self):
        test_api = "application/vnd.github.cloak-preview+json"
        user_agent = "brettcannon"
        headers = sansio.create_headers(user_agent, accept=test_api)
        assert headers["user-agent"] == user_agent
        assert headers["accept"] == test_api

    def test_all_keys_lowercase(self):
        """Test all header fields are lowercase."""
        user_agent = "brettcannon"
        test_api = "application/vnd.github.cloak-preview+json"
        oauth_token = "secret"
        headers = sansio.create_headers(
            user_agent, accept=test_api, oauth_token=oauth_token
        )
        assert len(headers) == 3
        for key in headers.keys():
            assert key == key.lower()
        assert headers["user-agent"] == user_agent
        assert headers["accept"] == test_api
        assert headers["authorization"] == f"token {oauth_token}"

    def test_authorization_with_jwt(self):
        user_agent = "brettcannon"
        jwt = "secret"
        headers = sansio.create_headers(user_agent, jwt=jwt)
        assert headers["user-agent"] == user_agent
        assert headers["authorization"] == f"bearer {jwt}"

    def test_cannot_pass_both_jwt_and_oauth(self):
        user_agent = "brettcannon"
        jwt = "secret jwt"
        oauth_token = "secret oauth token"
        with pytest.raises(ValueError) as exc_info:
            sansio.create_headers(user_agent, oauth_token=oauth_token, jwt=jwt)
        assert str(exc_info.value) == "Cannot pass both oauth_token and jwt."


class TestRateLimit:
    def test_init(self):
        left = 42
        rate = 64
        reset = datetime.datetime.now(datetime.timezone.utc)
        rate_limit = sansio.RateLimit(
            remaining=left, limit=rate, reset_epoch=reset.timestamp()
        )
        assert rate_limit.remaining == left
        assert rate_limit.limit == rate
        assert rate_limit.reset_datetime == reset

    def test_bool(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        year_from_now = now + datetime.timedelta(365)
        year_ago = now - datetime.timedelta(365)
        # Requests left.
        rate = sansio.RateLimit(
            remaining=1, limit=1, reset_epoch=year_from_now.timestamp()
        )
        assert rate
        # Reset passed.
        rate = sansio.RateLimit(remaining=0, limit=1, reset_epoch=year_ago.timestamp())
        assert rate
        # No requests and reset not passed.
        rate = sansio.RateLimit(
            remaining=0, limit=1, reset_epoch=year_from_now.timestamp()
        )
        assert not rate

    def test_from_http(self):
        left = 42
        rate = 65
        reset = datetime.datetime.now(datetime.timezone.utc)
        headers = {
            "x-ratelimit-limit": str(rate),
            "x-ratelimit-remaining": str(left),
            "x-ratelimit-reset": str(reset.timestamp()),
        }
        rate_limit = sansio.RateLimit.from_http(headers)
        assert rate_limit.limit == rate
        assert rate_limit.remaining == left
        assert rate_limit.reset_datetime == reset

    def test___str__(self):
        left = 4200
        rate = 65000
        reset = datetime.datetime.now(datetime.timezone.utc)
        message = str(
            sansio.RateLimit(limit=rate, remaining=left, reset_epoch=reset.timestamp())
        )
        assert format(left, ",") in message
        assert format(rate, ",") in message
        assert str(reset) in message

    def test_from_http_no_ratelimit(self):
        headers = {}
        rate_limit = sansio.RateLimit.from_http(headers)
        assert rate_limit is None


def sample(directory, status_code):
    # pytest doesn't set __spec__.origin :(
    sample_dir = pathlib.Path(__file__).parent / "samples" / directory
    headers_path = sample_dir / f"{status_code}.json"
    with headers_path.open("r") as file:
        headers = json.load(file)
    body = (sample_dir / "body").read_bytes()
    return headers, body


class TestDecipherResponse:
    """Tests for gidgethub.sansio.decipher_response()."""

    def test_5XX(self):
        status_code = 502
        with pytest.raises(GitHubBroken) as exc_info:
            sansio.decipher_response(status_code, {}, b"")
        assert exc_info.value.status_code == http.HTTPStatus(status_code)

    def test_4XX_no_message(self):
        status_code = 400
        with pytest.raises(BadRequest) as exc_info:
            sansio.decipher_response(status_code, {}, b"")
        assert exc_info.value.status_code == http.HTTPStatus(status_code)

    def test_4XX_message(self):
        status_code = 400
        message = json.dumps({"message": "it went bad"}).encode("UTF-8")
        headers = {"content-type": "application/json; charset=utf-8"}
        with pytest.raises(BadRequest) as exc_info:
            sansio.decipher_response(status_code, headers, message)
        assert exc_info.value.status_code == http.HTTPStatus(status_code)
        assert str(exc_info.value) == "it went bad"

    def test_404(self):
        status_code = 404
        headers, body = sample("pr_not_found", status_code)
        with pytest.raises(BadRequest) as exc_info:
            sansio.decipher_response(status_code, headers, body)
        assert exc_info.value.status_code == http.HTTPStatus(status_code)
        assert str(exc_info.value) == "Not Found"

    def test_403_rate_limit_exceeded(self):
        status_code = 403
        headers = {
            "content-type": "application/json; charset=utf-8",
            "x-ratelimit-limit": "2",
            "x-ratelimit-remaining": "0",
            "x-ratelimit-reset": "1",
        }
        body = json.dumps({"message": "oops"}).encode("UTF-8")
        with pytest.raises(RateLimitExceeded) as exc_info:
            sansio.decipher_response(status_code, headers, body)
        assert exc_info.value.status_code == http.HTTPStatus(status_code)

    def test_403_forbidden(self):
        status_code = 403
        headers = {
            "content-type": "application/json; charset=utf-8",
            "x-ratelimit-limit": "2",
            "x-ratelimit-remaining": "1",
            "x-ratelimit-reset": "1",
        }
        with pytest.raises(BadRequest) as exc_info:
            sansio.decipher_response(status_code, headers, b"")
        assert exc_info.value.status_code == http.HTTPStatus(status_code)

    def test_422(self):
        status_code = 422
        errors = [{"resource": "Issue", "field": "title", "code": "missing_field"}]
        body = json.dumps({"message": "it went bad", "errors": errors})
        body = body.encode("utf-8")
        headers = {"content-type": "application/json; charset=utf-8"}
        with pytest.raises(InvalidField) as exc_info:
            sansio.decipher_response(status_code, headers, body)
        assert exc_info.value.status_code == http.HTTPStatus(status_code)
        assert str(exc_info.value) == "it went bad for 'title'"

    def test_422_custom_code(self):
        status_code = 422
        errors = [
            {
                "resource": "PullRequest",
                "code": "custom",
                "message": "A pull request already exists for foo:1.",
            }
        ]
        body = json.dumps({"message": "it went bad", "errors": errors})
        body = body.encode("utf-8")
        headers = {"content-type": "application/json; charset=utf-8"}
        with pytest.raises(ValidationError) as exc_info:
            sansio.decipher_response(status_code, headers, body)
        assert exc_info.value.status_code == http.HTTPStatus(status_code)
        assert (
            str(exc_info.value)
            == "it went bad: 'A pull request already exists for foo:1.'"
        )

    def test_422_no_errors_object(self):
        status_code = 422
        body = json.dumps(
            {
                "message": "Reference does not exist",
                "documentation_url": "https://docs.github.com/en/free-pro-team@latest/rest/reference/git#delete-a-reference",
            }
        )
        body = body.encode("utf-8")
        headers = {"content-type": "application/json; charset=utf-8"}
        with pytest.raises(InvalidField) as exc_info:
            sansio.decipher_response(status_code, headers, body)
        assert exc_info.value.status_code == http.HTTPStatus(status_code)
        assert str(exc_info.value) == "Reference does not exist"

    def test_422_html_response(self):
        # https://github.com/brettcannon/gidgethub/issues/81
        status_code = 422
        body = "<html><body>Mistakes were made ...</body></html>"
        encoded_body = body.encode("utf-8")
        headers = {"content-type": "text/html; charset=utf-8"}
        with pytest.raises(BadRequestUnknownError) as exc_info:
            sansio.decipher_response(status_code, headers, encoded_body)
        assert exc_info.value.status_code == http.HTTPStatus(status_code)
        assert exc_info.value.response == body

    def test_3XX(self):
        status_code = 301
        with pytest.raises(RedirectionException) as exc_info:
            sansio.decipher_response(status_code, {}, b"")
        assert exc_info.value.status_code == http.HTTPStatus(status_code)

    def test_2XX_error(self):
        status_code = 205
        with pytest.raises(HTTPException) as exc_info:
            sansio.decipher_response(status_code, {}, b"")
        assert exc_info.value.status_code == http.HTTPStatus(status_code)

    def test_200(self):
        status_code = 200
        headers, body = sample("pr_single", status_code)
        data, rate_limit, more = sansio.decipher_response(status_code, headers, body)
        assert more is None
        assert rate_limit.remaining == 53
        assert data["url"] == "https://api.github.com/repos/python/cpython/pulls/1"

    def test_201(self):
        """Test a 201 response along with non-pagination Link header."""
        status_code = 201
        headers = {
            "x-ratelimit-limit": "60",
            "x-ratelimit-remaining": "50",
            "x-ratelimit-reset": "12345678",
            "content-type": "application/json; charset=utf-8",
            "link": '<http://example.com>; test="unimportant"',
        }
        data = {
            "id": 208_045_946,
            "url": "https://api.github.com/repos/octocat/Hello-World/labels/bug",
            "name": "bug",
            "color": "f29513",
            "default": True,
        }
        body = json.dumps(data).encode("UTF-8")
        returned_data, rate_limit, more = sansio.decipher_response(
            status_code, headers, body
        )
        assert more is None
        assert rate_limit.limit == 60
        assert returned_data == data

    def test_202(self):
        # https://github.com/brettcannon/gidgethub/issues/171
        status_code = 202
        headers = {
            "x-ratelimit-limit": "5000",
            "x-ratelimit-remaining": "4987",
            "x-ratelimit-reset": "1641847010",
            "content-type": "application/json; charset=utf-8",
        }
        data = {
            "id": 446_568_946,
            "name": "Hello-World",
            "fork": True,
            "forks": 0,
        }
        body = json.dumps(data).encode("UTF-8")
        returned_data, rate_limit, more = sansio.decipher_response(
            status_code, headers, body
        )
        assert more is None
        assert returned_data == data

    def test_204(self):
        """Test both a 204 response and an empty response body."""
        status_code = 204
        headers, body = sample("pr_merged", status_code)
        data, rate_limit, more = sansio.decipher_response(status_code, headers, body)
        assert more is None
        assert rate_limit.remaining == 41
        assert data is None

    def test_next(self):
        status_code = 200
        headers, body = sample("pr_page_1", status_code)
        data, rate_limit, more = sansio.decipher_response(status_code, headers, body)
        assert more == "https://api.github.com/repositories/4164482/pulls?page=2"
        assert rate_limit.remaining == 53
        assert data[0]["url"] == "https://api.github.com/repos/django/django/pulls/8053"

        headers, body = sample("pr_page_2", status_code)
        data, rate_limit, more = sansio.decipher_response(status_code, headers, body)
        assert more == "https://api.github.com/repositories/4164482/pulls?page=3"
        assert rate_limit.remaining == 50
        assert data[0]["url"] == "https://api.github.com/repos/django/django/pulls/7805"

        headers, body = sample("pr_page_last", status_code)
        data, rate_limit, more = sansio.decipher_response(status_code, headers, body)
        assert more is None
        assert rate_limit.remaining == 48
        assert data[0]["url"] == "https://api.github.com/repos/django/django/pulls/6395"

    @pytest.mark.asyncio
    def test_next_with_search_api(self):
        status_code = 200
        headers, body = sample("search_issues_page_1", status_code)
        data, rate_limit, more = sansio.decipher_response(status_code, headers, body)
        assert more == (
            "https://api.github.com/search/issues"
            "?q=repo%3Abrettcannon%2Fgidgethub+state%3Aclosed"
            "+rate+&per_page=3&page=2"
        )
        assert rate_limit.remaining == 9
        assert {"items", "incomplete_results", "total_count"} == data.keys()
        expected_first_url = (
            "https://api.github.com/repos/brettcannon/gidgethub/issues/25"
        )
        assert data["items"][0]["url"] == expected_first_url

        headers, body = sample("search_issues_page_last", status_code)
        data, rate_limit, more = sansio.decipher_response(status_code, headers, body)
        assert more is None
        assert rate_limit.remaining == 9
        assert {"items", "incomplete_results", "total_count"} == data.keys()
        expected_first_url = (
            "https://api.github.com/repos/brettcannon/gidgethub/issues/10"
        )
        assert data["items"][0]["url"] == expected_first_url

    def test_text_body(self):
        """Test requesting non-JSON data like a diff."""
        status_code = 200
        headers, body = sample("pr_diff", status_code)
        data, rate_limit, more = sansio.decipher_response(status_code, headers, body)
        assert more is None
        assert rate_limit.remaining == 43
        assert data.startswith("diff --git")

    def test_no_ratelimit(self):
        """Test no ratelimit in headers."""
        status_code = 201
        headers = {
            "content-type": "application/json; charset=utf-8",
            "link": '<http://example.com>; test="unimportant"',
        }
        data = {
            "id": 208_045_946,
            "url": "https://api.github.com/repos/octocat/Hello-World/labels/bug",
            "name": "bug",
            "color": "f29513",
            "default": True,
        }
        body = json.dumps(data).encode("UTF-8")
        returned_data, rate_limit, more = sansio.decipher_response(
            status_code, headers, body
        )
        assert more is None
        assert rate_limit is None
        assert returned_data == data


class TestFormatUrl:
    """Tests for gidgethub.sansio.format_url()."""

    @pytest.mark.parametrize(
        "base_url",
        ["https://api.github.com/notifications", "https://my.host.com/notifications"],
    )
    def test_absolute_url(self, base_url):
        url = sansio.format_url(base_url, {}, base_url=base_url)
        assert url == base_url

    def test_different_base_and_absolute_url(self):
        url = sansio.format_url(
            "https://api.github.com/notifications",
            {},
            base_url="https://my.host.com/notifications",
        )
        assert url == "https://api.github.com/notifications"

    @pytest.mark.parametrize(
        "base_url", ["https://api.github.com", "https://my.host.com"]
    )
    def test_relative_url(self, base_url):
        url = sansio.format_url("/notifications", {}, base_url=base_url)
        assert url == f"{base_url}/notifications"

    def test_template(self):
        template_url = "https://api.github.com/users/octocat/gists{/gist_id}"
        template_data = {"gist_id": "1234"}
        # Substituting an absolute URL.
        url = sansio.format_url(template_url, template_data)
        assert url == "https://api.github.com/users/octocat/gists/1234"
        # No substituting an absolute URL.
        url = sansio.format_url(template_url, {})
        assert url == "https://api.github.com/users/octocat/gists"
        # Substituting a relative URL.
        url = sansio.format_url("/users/octocat/gists{/gist_id}", template_data)
        assert url == "https://api.github.com/users/octocat/gists/1234"

    def test_template_with_base_url(self):
        template_url = "https://my.host.com/users/octocat/gists{/gist_id}"
        template_data = {"gist_id": "1234"}
        # Substituting an absolute URL.
        url = sansio.format_url(
            template_url, template_data, base_url="https://my.host.com"
        )
        assert url == "https://my.host.com/users/octocat/gists/1234"
        # No substituting an absolute URL.
        url = sansio.format_url(template_url, {}, base_url="https://my.host.com")
        assert url == "https://my.host.com/users/octocat/gists"
        # Substituting a relative URL.
        url = sansio.format_url(
            "/users/octocat/gists{/gist_id}",
            template_data,
            base_url="https://my.host.com",
        )
        assert url == "https://my.host.com/users/octocat/gists/1234"

    @pytest.mark.parametrize(
        "base_url", ["https://api.github.com", "https://my.host.com"]
    )
    def test_quoting(self, base_url):
        template_url = "https://api.github.com/repos/python/cpython/labels{/name}"
        label = {"name": "CLA signed"}
        url = sansio.format_url(template_url, label, base_url=base_url)
        assert url == "https://api.github.com/repos/python/cpython/labels/CLA%20signed"



from unittest import mock

import importlib_resources
import jwt
import pytest

from gidgethub import apps
from .test_abc import MockGitHubAPI

from .samples import rsa_key as rsa_key_samples


class TestGitHubAppUtils:
    """Tests for GitHub App utilities."""

    @mock.patch("time.time")
    def test_get_jwt(self, time_mock):
        app_id = 12345

        time_mock.return_value = 1587069751.5588422

        # test file copied from https://github.com/jpadilla/pyjwt/blob/master/tests/keys/testkey_rsa
        private_key = (
            importlib_resources.files(rsa_key_samples) / "test_rsa_key"
        ).read_bytes()

        result = apps.get_jwt(app_id=app_id, private_key=private_key)
        expected_payload = {
            "iat": 1587069751,
            "exp": 1587069751 + (10 * 60),
            "iss": app_id,
        }

        assert result == jwt.encode(expected_payload, private_key, algorithm="RS256")

    @mock.patch("time.time")
    def test_get_jwt_with_custom_expiry(self, time_mock):
        app_id = 12345

        time_mock.return_value = 1587069751.5588422

        # test file copied from https://github.com/jpadilla/pyjwt/blob/master/tests/keys/testkey_rsa
        private_key = (
            importlib_resources.files(rsa_key_samples) / "test_rsa_key"
        ).read_bytes()

        # Custom expiration
        expiration = 30 * 60
        result = apps.get_jwt(
            app_id=app_id, private_key=private_key, expiration=expiration
        )
        expected_payload = {
            "iat": 1587069751,
            "exp": 1587069751 + expiration,
            "iss": app_id,
        }

        assert result == jwt.encode(expected_payload, private_key, algorithm="RS256")

    @pytest.mark.asyncio
    async def test_get_installation_access_token(self):
        gh = MockGitHubAPI()
        installation_id = 6789
        app_id = 12345

        private_key = (
            importlib_resources.files(rsa_key_samples) / "test_rsa_key"
        ).read_bytes()

        await apps.get_installation_access_token(
            gh, installation_id=installation_id, app_id=app_id, private_key=private_key
        )

        assert gh.url == "https://api.github.com/app/installations/6789/access_tokens"
        assert gh.body == b""



from typing import Any, Awaitable, Callable, Dict, List, FrozenSet

from . import sansio


AsyncCallback = Callable[..., Awaitable[None]]


class Router:
    """Route webhook events to registered functions."""

    def __init__(self, *other_routers: "Router") -> None:
        """Instantiate a new router (possibly from other routers)."""
        self._shallow_routes: Dict[str, List[AsyncCallback]] = {}
        # event type -> data key -> data value -> callbacks
        self._deep_routes: Dict[str, Dict[str, Dict[Any, List[AsyncCallback]]]] = {}
        for other_router in other_routers:
            for event_type, callbacks in other_router._shallow_routes.items():
                for callback in callbacks:
                    self.add(callback, event_type)
            for event_type, data_details in other_router._deep_routes.items():
                for data_key, data_specifics in data_details.items():
                    for data_value, callbacks in data_specifics.items():
                        detail = {data_key: data_value}
                        for callback in callbacks:
                            self.add(callback, event_type, **detail)

    def add(self, func: AsyncCallback, event_type: str, **data_detail: Any) -> None:
        """Add a new route.

        After registering 'func' for the specified event_type, an
        optional data_detail may be provided. By providing an extra
        keyword argument, dispatching can occur based on a top-level
        key of the data in the event being dispatched.
        """
        if len(data_detail) > 1:
            raise TypeError(
                "dispatching based on data details is only "
                "supported up to one level deep; "
                f"{len(data_detail)} levels specified"
            )
        elif not data_detail:
            callbacks = self._shallow_routes.setdefault(event_type, [])
            callbacks.append(func)
        else:
            data_key, data_value = data_detail.popitem()
            data_details = self._deep_routes.setdefault(event_type, {})
            specific_detail = data_details.setdefault(data_key, {})
            callbacks = specific_detail.setdefault(data_value, [])
            callbacks.append(func)

    def register(
        self, event_type: str, **data_detail: Any
    ) -> Callable[[AsyncCallback], AsyncCallback]:
        """Decorator to apply the add() method to a function."""

        def decorator(func: AsyncCallback) -> AsyncCallback:
            self.add(func, event_type, **data_detail)
            return func

        return decorator

    def fetch(self, event: sansio.Event) -> FrozenSet[AsyncCallback]:
        """Return a set of function(s) registered to the router that the event would
        be called on."""
        found_callbacks = set()
        try:
            found_callbacks.update(self._shallow_routes[event.event])
        except KeyError:
            pass
        try:
            details = self._deep_routes[event.event]
        except KeyError:
            pass
        else:
            for data_key, data_values in details.items():
                if data_key in event.data:
                    event_value = event.data[data_key]
                    if event_value in data_values:
                        found_callbacks.update(data_values[event_value])
        return frozenset(found_callbacks)

    async def dispatch(self, event: sansio.Event, *args: Any, **kwargs: Any) -> None:
        """Dispatch an event to all registered function(s)."""

        found_callbacks = self.fetch(event)
        for callback in found_callbacks:
            await callback(event, *args, **kwargs)



import json
import os

import pytest

from gidgethub import actions
from gidgethub.actions import _DELIMITER


@pytest.fixture
def tmp_webhook(tmp_path, monkeypatch):
    """Create a temporary file for an actions webhook event."""
    tmp_file_path = tmp_path / "event.json"
    monkeypatch.setenv("GITHUB_EVENT_PATH", os.fspath(tmp_file_path))
    return tmp_file_path


@pytest.fixture
def tmp_envfile(tmp_path, monkeypatch):
    """Create a temporary environment file."""
    tmp_file_path = tmp_path / "setenv.txt"
    monkeypatch.setenv("GITHUB_ENV", os.fspath(tmp_file_path))
    return tmp_file_path


@pytest.fixture
def tmp_pathfile(tmpdir, tmp_path, monkeypatch):
    """Create a temporary system path file and a temporary directory path."""
    tmp_file_path = tmp_path / "addpath.txt"
    monkeypatch.setenv("GITHUB_PATH", os.fspath(tmp_file_path))
    return tmp_file_path, tmpdir


class TestWorkspace:
    """Tests for gidgethub.actions.workspace()."""

    def test_reading(self, tmp_path, monkeypatch):
        monkeypatch.setenv("GITHUB_WORKSPACE", os.fspath(tmp_path))
        assert actions.workspace() == tmp_path

    def test_caching(self, tmp_path, monkeypatch):
        actions.workspace.cache_clear()
        monkeypatch.setenv("GITHUB_WORKSPACE", os.fspath(tmp_path))
        assert actions.workspace() == tmp_path
        new_path = tmp_path / "DOES NOT EXIST"
        monkeypatch.setenv("GITHUB_WORKSPACE", os.fspath(new_path))
        assert actions.workspace() == tmp_path


class TestEvent:
    """Tests for gidgethub.actions.event()."""

    def test_reading(self, tmp_webhook):
        data = {"something": "going on"}
        tmp_webhook.write_text(json.dumps(data), encoding="utf-8")
        assert actions.event() == data

    def test_caching(self, tmp_webhook):
        actions.event.cache_clear()
        self.test_reading(tmp_webhook)
        original_data = actions.event()
        new_data = {"something": "completely different"}
        tmp_webhook.write_text(json.dumps(new_data), encoding="utf-8")
        assert actions.event() == original_data


class TestCommand:
    """Tests for gidgethub.actions.command()."""

    # https://help.github.com/en/actions/reference/development-tools-for-github-actions#logging-commands

    def _stdout(self, capsys):
        stdout, stderr = capsys.readouterr()
        assert not stderr
        assert stdout
        return stdout.strip()

    def test_set_output(self, capsys):
        actions.command("set-output", "strawberry", name="action_fruit")
        assert self._stdout(capsys) == "::set-output name=action_fruit::strawberry"

    def test_debug(self, capsys):
        actions.command(
            "debug", "Entered octocatAddition method", file="app.js", line="1"
        )
        assert (
            self._stdout(capsys)
            == "::debug file=app.js,line=1::Entered octocatAddition method"
        )

    def test_warning(self, capsys):
        actions.command(
            "warning", "Missing semicolon", file="app.js", line="1", col="5"
        )
        assert (
            self._stdout(capsys)
            == "::warning file=app.js,line=1,col=5::Missing semicolon"
        )

    def test_error(self, capsys):
        actions.command(
            "error", "Something went wrong", file="app.js", line="10", col="15"
        )
        assert (
            self._stdout(capsys)
            == "::error file=app.js,line=10,col=15::Something went wrong"
        )

    def test_add_mask(self, capsys):
        actions.command("add-mask", "Mona The Octocat")
        assert self._stdout(capsys) == "::add-mask::Mona The Octocat"

    def test_stop_commands(self, capsys):
        actions.command("stop-commands", "pause-logging")
        assert self._stdout(capsys) == "::stop-commands::pause-logging"

    def test_resume_command(self, capsys):
        actions.command("pause-logging")
        assert self._stdout(capsys) == "::pause-logging::"


class TestSetenv:
    """Tests for gidgethub.actions.setenv()."""

    def test_creating(self, tmp_envfile):
        actions.setenv("HELLO", "WORLD")
        data = tmp_envfile.read_text(encoding="utf-8")
        assert os.environ["HELLO"] == "WORLD"
        assert (
            data == f"HELLO<<{_DELIMITER}{os.linesep}WORLD{os.linesep}"
            f"{_DELIMITER}{os.linesep}"
        )

    def test_updating(self, tmp_envfile):
        actions.setenv("CHANGED", "FALSE")
        data = tmp_envfile.read_text(encoding="utf-8")
        assert os.environ["CHANGED"] == "FALSE"
        assert (
            data == f"CHANGED<<{_DELIMITER}{os.linesep}FALSE{os.linesep}"
            f"{_DELIMITER}{os.linesep}"
        )
        actions.setenv("CHANGED", "TRUE")
        updated = tmp_envfile.read_text(encoding="utf-8")
        assert os.environ["CHANGED"] == "TRUE"
        # Rendering of the updated variable is done by GitHub.
        assert updated.endswith(
            f"CHANGED<<{_DELIMITER}{os.linesep}TRUE{os.linesep}"
            f"{_DELIMITER}{os.linesep}"
        )

    def test_creating_multiline(self, tmp_envfile):
        multiline = """This
                        is
                        a
                        multiline
                        string."""
        actions.setenv("MULTILINE", multiline)
        data = tmp_envfile.read_text(encoding="utf-8")
        assert os.environ["MULTILINE"] == multiline
        assert (
            data == f"MULTILINE<<{_DELIMITER}{os.linesep}{multiline}"
            f"{os.linesep}{_DELIMITER}{os.linesep}"
        )


class TestAddpath:
    """Tests for gidgethub.actions.addpath()."""

    def test_string_path(self, tmp_pathfile):
        actions.addpath("/path/to/random/dir")
        data = tmp_pathfile[0].read_text(encoding="utf-8")
        assert f"/path/to/random/dir{os.pathsep}" in os.environ["PATH"]
        assert data == f"/path/to/random/dir{os.linesep}"

    def test_path_object(self, tmp_pathfile):
        actions.addpath(tmp_pathfile[1])
        data = tmp_pathfile[0].read_text(encoding="utf-8")
        assert f"{tmp_pathfile[1]!s}{os.pathsep}" in os.environ["PATH"]
        assert data == f"{tmp_pathfile[1]!s}{os.linesep}"

    def test_multiple_paths(self, tmp_pathfile):
        actions.addpath("/path/to/random/dir")
        random_path = tmp_pathfile[1] / "random.txt"
        actions.addpath(random_path)
        data = tmp_pathfile[0].read_text(encoding="utf-8")
        # Last path added comes first.
        assert (
            f"{random_path!s}{os.pathsep}/path/to/random/dir{os.pathsep}"
            in os.environ["PATH"]
        )
        assert data == f"/path/to/random/dir{os.linesep}{random_path!s}{os.linesep}"



import asyncio
from typing import Mapping, Tuple, Any

import httpx

from . import abc as gh_abc


class GitHubAPI(gh_abc.GitHubAPI):
    def __init__(self, client: httpx.AsyncClient, *args: Any, **kwargs: Any) -> None:
        self._client = client
        super().__init__(*args, **kwargs)

    async def _request(
        self, method: str, url: str, headers: Mapping[str, str], body: bytes = b""
    ) -> Tuple[int, Mapping[str, str], bytes]:
        """Make an HTTP request."""
        response = await self._client.request(
            method, url, headers=dict(headers), content=body
        )
        return response.status_code, response.headers, response.content

    async def sleep(self, seconds: float) -> None:
        """Sleep for the specified number of seconds."""
        await asyncio.sleep(seconds)



import http
import json
import re

import importlib_resources
import pytest

from gidgethub import (
    BadGraphQLRequest,
    GitHubBroken,
    GraphQLAuthorizationFailure,
    GraphQLException,
    GraphQLResponseTypeError,
    QueryError,
    RedirectionException,
    sansio,
)
from gidgethub import abc as gh_abc
from gidgethub.abc import JSON_UTF_8_CHARSET

from .samples import GraphQL as graphql_samples


class MockGitHubAPI(gh_abc.GitHubAPI):
    DEFAULT_HEADERS = {
        "x-ratelimit-limit": "2",
        "x-ratelimit-remaining": "1",
        "x-ratelimit-reset": "0",
        "content-type": JSON_UTF_8_CHARSET,
    }

    def __init__(
        self,
        status_code=200,
        headers=DEFAULT_HEADERS,
        body=b"",
        *,
        cache=None,
        oauth_token=None,
        base_url=sansio.DOMAIN,
    ):
        self.response_code = status_code
        self.response_headers = headers
        self.response_body = body
        super().__init__(
            "test_abc", oauth_token=oauth_token, cache=cache, base_url=base_url
        )

    async def _request(self, method, url, headers, body=b""):
        """Make an HTTP request."""
        self.method = method
        self.url = url
        self.headers = headers
        self.body = body
        response_headers = self.response_headers.copy()
        try:
            # Don't loop forever.
            del self.response_headers["link"]
        except KeyError:
            pass
        return self.response_code, response_headers, self.response_body

    async def sleep(self, seconds):  # pragma: no cover
        """Sleep for the specified number of seconds."""
        self.slept = seconds


class TestGeneralGitHubAPI:
    @pytest.mark.asyncio
    async def test_url_formatted(self):
        """The URL is appropriately formatted."""
        gh = MockGitHubAPI()
        await gh._make_request(
            "GET",
            "/users/octocat/following{/other_user}",
            {"other_user": "brettcannon"},
            "",
            sansio.accept_format(),
        )
        assert gh.url == "https://api.github.com/users/octocat/following/brettcannon"

    @pytest.mark.asyncio
    async def test_url_formatted_with_base_url(self):
        """The URL is appropriately formatted."""
        gh = MockGitHubAPI(base_url="https://my.host.com")
        await gh._make_request(
            "GET",
            "/users/octocat/following{/other_user}",
            {"other_user": "brettcannon"},
            "",
            sansio.accept_format(),
        )
        assert gh.url == "https://my.host.com/users/octocat/following/brettcannon"

    @pytest.mark.asyncio
    async def test_headers(self):
        """Appropriate headers are created."""
        accept = sansio.accept_format()
        gh = MockGitHubAPI(oauth_token="oauth token")
        await gh._make_request("GET", "/rate_limit", {}, "", accept)
        assert gh.headers["user-agent"] == "test_abc"
        assert gh.headers["accept"] == accept
        assert gh.headers["authorization"] == "token oauth token"

    @pytest.mark.asyncio
    async def test_auth_headers_with_passed_token(self):
        """Test the authorization header with the passed oauth_token."""
        accept = sansio.accept_format()
        gh = MockGitHubAPI()
        await gh._make_request(
            "GET", "/rate_limit", {}, "", accept, oauth_token="oauth token"
        )
        assert gh.headers["user-agent"] == "test_abc"
        assert gh.headers["accept"] == accept
        assert gh.headers["authorization"] == "token oauth token"

    @pytest.mark.asyncio
    async def test_auth_headers_with_passed_jwt(self):
        """Test the authorization header with the passed jwt."""
        accept = sansio.accept_format()
        gh = MockGitHubAPI()
        await gh._make_request(
            "GET", "/rate_limit", {}, "", accept, jwt="json web token"
        )
        assert gh.headers["user-agent"] == "test_abc"
        assert gh.headers["accept"] == accept
        assert gh.headers["authorization"] == "bearer json web token"

    @pytest.mark.asyncio
    async def test_make_request_passing_token_and_jwt(self):
        """Test that passing both jwt and oauth_token raises ValueError."""
        accept = sansio.accept_format()
        gh = MockGitHubAPI()
        with pytest.raises(ValueError) as exc_info:
            await gh._make_request(
                "GET",
                "/rate_limit",
                {},
                "",
                accept,
                jwt="json web token",
                oauth_token="oauth token",
            )
        assert str(exc_info.value) == "Cannot pass both oauth_token and jwt."

    @pytest.mark.asyncio
    async def test_rate_limit_set(self):
        """The rate limit is updated after receiving a response."""
        rate_headers = {
            "x-ratelimit-limit": "42",
            "x-ratelimit-remaining": "1",
            "x-ratelimit-reset": "0",
        }
        gh = MockGitHubAPI(headers=rate_headers)
        await gh._make_request("GET", "/rate_limit", {}, "", sansio.accept_format())
        assert gh.rate_limit.limit == 42

    @pytest.mark.asyncio
    async def test_decoding(self):
        """Test that appropriate decoding occurs."""
        original_data = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        data, _, _ = await gh._make_request(
            "GET", "/rate_limit", {}, "", sansio.accept_format()
        )
        assert data == original_data

    @pytest.mark.asyncio
    async def test_more(self):
        """The 'next' link is returned appropriately."""
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["link"] = "<https://api.github.com/fake?page=2>; " 'rel="next"'
        gh = MockGitHubAPI(headers=headers)
        _, more, _ = await gh._make_request(
            "GET", "/fake", {}, "", sansio.accept_format()
        )
        assert more == "https://api.github.com/fake?page=2"

    @pytest.mark.asyncio
    async def test_status_code(self):
        """The status code is returned appropriately."""
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["link"] = "<https://api.github.com/fake?page=2>; " 'rel="next"'
        gh = MockGitHubAPI(headers=headers)
        _, _, status_code = await gh._make_request(
            "GET", "/fake", {}, "", sansio.accept_format()
        )
        assert status_code == 200


class TestGitHubAPIGetitem:
    @pytest.mark.asyncio
    async def test_getitem(self):
        original_data = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        data = await gh.getitem("/fake")
        assert gh.method == "GET"
        assert data == original_data

    @pytest.mark.asyncio
    async def test_with_passed_jwt(self):
        original_data = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        await gh.getitem("/fake", jwt="json web token")
        assert gh.method == "GET"
        assert gh.headers["authorization"] == "bearer json web token"

    @pytest.mark.asyncio
    async def test_with_passed_oauth_token(self):
        original_data = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        await gh.getitem("/fake", oauth_token="my oauth token")
        assert gh.method == "GET"
        assert gh.headers["authorization"] == "token my oauth token"

    @pytest.mark.asyncio
    async def test_cannot_pass_both_jwt_and_oauth_token(self):
        original_data = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        with pytest.raises(ValueError) as exc_info:
            await gh.getitem(
                "/fake", oauth_token="my oauth token", jwt="json web token"
            )

        assert str(exc_info.value) == "Cannot pass both oauth_token and jwt."


class TestGitHubAPIGetStatus:
    @pytest.mark.asyncio
    async def test_getstatus(self):
        original_status = 204
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        gh = MockGitHubAPI(headers=headers, status_code=original_status)
        data = await gh.getstatus("/fake")
        assert gh.method == "GET"
        assert data == original_status

    @pytest.mark.asyncio
    async def test_getstatus_4xx(self):
        original_status = 404
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        gh = MockGitHubAPI(headers=headers, status_code=original_status)
        data = await gh.getstatus("/fake")
        assert gh.method == "GET"
        assert data == original_status

    @pytest.mark.asyncio
    async def test_getstatus_5xx(self):
        original_status = 504
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        gh = MockGitHubAPI(headers=headers, status_code=original_status)
        data = await gh.getstatus("/fake")
        assert gh.method == "GET"
        assert data == original_status


class TestGitHubAPIGetiter:
    @pytest.mark.asyncio
    async def test_getiter(self):
        """Test that getiter() returns an async iterable as well as URI expansion."""
        original_data = [1, 2]
        next_url = "https://api.github.com/fake{/extra}?page=2"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        headers["link"] = f'<{next_url}>; rel="next"'
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        data = []
        async for item in gh.getiter("/fake", {"extra": "stuff"}):
            data.append(item)
        assert gh.method == "GET"
        assert gh.url == "https://api.github.com/fake/stuff?page=2"
        assert len(data) == 4
        assert data[0] == 1
        assert data[1] == 2
        assert data[2] == 1
        assert data[3] == 2

    @pytest.mark.asyncio
    async def test_getiter_with_extra_headers(self):
        """Test that getiter() sends extra headers correctly."""
        original_data = [1, 2]
        next_url = "https://api.github.com/fake{/extra}?page=2"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        headers["link"] = f'<{next_url}>; rel="next"'
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        extra_headers = {"X-Custom-Header": "custom_value"}
        data = []
        async for item in gh.getiter(
            "/fake", {"extra": "stuff"}, extra_headers=extra_headers
        ):
            data.append(item)
        assert gh.method == "GET"
        assert gh.headers["X-Custom-Header"] == "custom_value"
        assert len(data) == 4
        assert data[0] == 1
        assert data[1] == 2
        assert data[2] == 1
        assert data[3] == 2

    @pytest.mark.asyncio
    async def test_with_passed_jwt(self):
        original_data = [1, 2]
        next_url = "https://api.github.com/fake{/extra}?page=2"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        headers["link"] = f'<{next_url}>; rel="next"'
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        data = []
        async for item in gh.getiter("/fake", {"extra": "stuff"}, jwt="json web token"):
            data.append(item)
        assert gh.method == "GET"
        assert gh.headers["authorization"] == "bearer json web token"

    @pytest.mark.asyncio
    async def test_with_passed_oauth_token(self):
        original_data = [1, 2]
        next_url = "https://api.github.com/fake{/extra}?page=2"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        headers["link"] = f'<{next_url}>; rel="next"'
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        data = []
        async for item in gh.getiter(
            "/fake", {"extra": "stuff"}, oauth_token="my oauth token"
        ):
            data.append(item)
        assert gh.method == "GET"
        assert gh.headers["authorization"] == "token my oauth token"

    @pytest.mark.asyncio
    async def test_cannot_pass_both_oauth_and_jwt(self):
        original_data = [1, 2]
        next_url = "https://api.github.com/fake{/extra}?page=2"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        headers["link"] = f'<{next_url}>; rel="next"'
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        with pytest.raises(ValueError) as exc_info:
            async for _ in gh.getiter(
                "/fake",
                {"extra": "stuff"},
                oauth_token="my oauth token",
                jwt="json web token",
            ):
                pytest.fail("Unreachable")  # pragma: no cover
        assert str(exc_info.value) == "Cannot pass both oauth_token and jwt."

    @pytest.mark.asyncio
    async def test_search_api(self):
        """Test that getiter() returns an async iterable with the items if
        the original data is a dictionary (GitHub search API)."""
        original_data = {"items": [1, 2]}
        next_url = "https://api.github.com/fake{/extra}?page=2"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        headers["link"] = f'<{next_url}>; rel="next"'
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        data = []
        async for item in gh.getiter("/fake", {"extra": "stuff"}):
            data.append(item)
        assert gh.method == "GET"
        assert gh.url == "https://api.github.com/fake/stuff?page=2"
        assert len(data) == 4
        assert data[0] == 1
        assert data[1] == 2
        assert data[2] == 1
        assert data[3] == 2

    @pytest.mark.asyncio
    async def test_checks_api(self):
        """Same as search API, but using the `checks_runs` parameter instead of `items`."""
        original_data = {"check_runs": [1, 2]}
        next_url = "https://api.github.com/fake{/extra}?page=2"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=UTF-8"
        headers["link"] = f'<{next_url}>; rel="next"'
        gh = MockGitHubAPI(
            headers=headers, body=json.dumps(original_data).encode("utf8")
        )
        data = []
        async for item in gh.getiter(
            "/fake", {"extra": "stuff"}, iterable_key="check_runs"
        ):
            data.append(item)
        assert gh.method == "GET"
        assert gh.url == "https://api.github.com/fake/stuff?page=2"
        assert len(data) == 4
        assert data[0] == 1
        assert data[1] == 2
        assert data[2] == 1
        assert data[3] == 2

    @pytest.mark.asyncio
    async def test_extra_headers(self):
        """Test that extra headers are passed correctly."""
        accept = sansio.accept_format()
        extra_headers = {"X-Custom-Header": "custom_value"}
        gh = MockGitHubAPI()
        await gh._make_request(
            "GET", "/rate_limit", {}, "", accept, extra_headers=extra_headers
        )
        assert gh.headers["user-agent"] == "test_abc"
        assert gh.headers["accept"] == accept
        assert "X-Custom-Header" in gh.headers
        assert gh.headers["X-Custom-Header"] == "custom_value"


class TestGitHubAPIPost:
    @pytest.mark.asyncio
    async def test_post(self):
        send = [1, 2, 3]
        send_json = json.dumps(send).encode("utf-8")
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.post("/fake", data=send)
        assert gh.method == "POST"
        assert gh.headers["content-type"] == "application/json; charset=utf-8"
        assert gh.body == send_json
        assert gh.headers["content-length"] == str(len(send_json))
        assert gh.headers["content-type"] == JSON_UTF_8_CHARSET

    @pytest.mark.asyncio
    async def test_with_passed_jwt(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.post("/fake", data=send, jwt="json web token")
        assert gh.method == "POST"
        assert gh.headers["authorization"] == "bearer json web token"
        assert gh.headers["content-type"] == JSON_UTF_8_CHARSET

    @pytest.mark.asyncio
    async def test_with_passed_oauth_token(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.post("/fake", data=send, oauth_token="my oauth token")
        assert gh.method == "POST"
        assert gh.headers["authorization"] == "token my oauth token"
        assert gh.headers["content-type"] == JSON_UTF_8_CHARSET

    @pytest.mark.asyncio
    async def test_cannot_pass_both_oauth_and_jwt(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        with pytest.raises(ValueError) as exc_info:
            await gh.post(
                "/fake", data=send, oauth_token="my oauth token", jwt="json web token"
            )

        assert str(exc_info.value) == "Cannot pass both oauth_token and jwt."

    @pytest.mark.asyncio
    async def test_with_passed_content_type(self):
        """Assert that the body is not parsed to JSON and the content-type header is set."""
        gh = MockGitHubAPI()
        data = "blabla"
        await gh.post("/fake", data=data, content_type="application/zip")
        assert gh.method == "POST"
        assert gh.headers["content-type"] == "application/zip"
        assert gh.body == data
        assert gh.headers["content-length"] == str(len(data))


class TestGitHubAPIPatch:
    @pytest.mark.asyncio
    async def test_patch(self):
        send = [1, 2, 3]
        send_json = json.dumps(send).encode("utf-8")
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.patch("/fake", data=send)
        assert gh.method == "PATCH"
        assert gh.headers["content-type"] == "application/json; charset=utf-8"
        assert gh.body == send_json
        assert gh.headers["content-length"] == str(len(send_json))

    @pytest.mark.asyncio
    async def test_with_passed_jwt(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.patch("/fake", data=send, jwt="json web token")
        assert gh.method == "PATCH"
        assert gh.headers["authorization"] == "bearer json web token"

    @pytest.mark.asyncio
    async def test_with_passed_oauth_token(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.patch("/fake", data=send, oauth_token="my oauth token")
        assert gh.method == "PATCH"
        assert gh.headers["authorization"] == "token my oauth token"

    @pytest.mark.asyncio
    async def test_cannot_pass_both_oauth_and_jwt(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        with pytest.raises(ValueError) as exc_info:
            await gh.patch(
                "/fake", data=send, oauth_token="my oauth token", jwt="json web token"
            )

        assert str(exc_info.value) == "Cannot pass both oauth_token and jwt."


class TestGitHubAPIPut:
    @pytest.mark.asyncio
    async def test_put(self):
        send = [1, 2, 3]
        send_json = json.dumps(send).encode("utf-8")
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.put("/fake", data=send)
        assert gh.method == "PUT"
        assert gh.headers["content-type"] == "application/json; charset=utf-8"
        assert gh.body == send_json
        assert gh.headers["content-length"] == str(len(send_json))

    @pytest.mark.asyncio
    async def test_with_passed_jwt(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.put("/fake", data=send, jwt="json web token")
        assert gh.method == "PUT"
        assert gh.headers["authorization"] == "bearer json web token"

    @pytest.mark.asyncio
    async def test_with_passed_oauth_token(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.put("/fake", data=send, oauth_token="my oauth token")
        assert gh.method == "PUT"
        assert gh.headers["authorization"] == "token my oauth token"

    @pytest.mark.asyncio
    async def test_cannot_pass_both_oauth_and_jwt(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        with pytest.raises(ValueError) as exc_info:
            await gh.put(
                "/fake", data=send, oauth_token="my oauth token", jwt="json web token"
            )

        assert str(exc_info.value) == "Cannot pass both oauth_token and jwt."


class TestGitHubAPIDelete:
    @pytest.mark.asyncio
    async def test_delete(self):
        send = [1, 2, 3]
        send_json = json.dumps(send).encode("utf-8")
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.delete("/fake", data=send)
        assert gh.method == "DELETE"
        assert gh.headers["content-type"] == "application/json; charset=utf-8"
        assert gh.body == send_json
        assert gh.headers["content-length"] == str(len(send_json))

    @pytest.mark.asyncio
    async def test_with_passed_jwt(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.delete("/fake", data=send, jwt="json web token")
        assert gh.method == "DELETE"
        assert gh.headers["authorization"] == "bearer json web token"

    @pytest.mark.asyncio
    async def test_with_passed_oauth_token(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        await gh.delete("/fake", data=send, oauth_token="my oauth token")
        assert gh.method == "DELETE"
        assert gh.headers["authorization"] == "token my oauth token"

    @pytest.mark.asyncio
    async def test_pass_both_oauth_and_jwt(self):
        send = [1, 2, 3]
        receive = {"hello": "world"}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["content-type"] = "application/json; charset=utf-8"
        gh = MockGitHubAPI(headers=headers, body=json.dumps(receive).encode("utf-8"))
        with pytest.raises(ValueError) as exc_info:
            await gh.delete(
                "/fake", data=send, oauth_token="my oauth token", jwt="json web token"
            )

        assert str(exc_info.value) == "Cannot pass both oauth_token and jwt."


class TestGitHubAPICache:
    @pytest.mark.asyncio
    async def test_if_none_match_sent(self):
        etag = "12345"
        cache = {"https://api.github.com/fake": (etag, None, "hi", None)}
        gh = MockGitHubAPI(cache=cache)
        await gh.getitem("/fake")
        assert "if-none-match" in gh.headers
        assert gh.headers["if-none-match"] == etag

    @pytest.mark.asyncio
    async def test_etag_received(self):
        cache = {}
        etag = "12345"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["etag"] = etag
        gh = MockGitHubAPI(200, headers, b"42", cache=cache)
        data = await gh.getitem("/fake")
        url = "https://api.github.com/fake"
        assert url in cache
        assert cache[url] == (etag, None, 42, None)
        assert data == cache[url][2]

    @pytest.mark.asyncio
    async def test_if_modified_since_sent(self):
        last_modified = "12345"
        cache = {"https://api.github.com/fake": (None, last_modified, "hi", None)}
        gh = MockGitHubAPI(cache=cache)
        await gh.getitem("/fake")
        assert "if-modified-since" in gh.headers
        assert gh.headers["if-modified-since"] == last_modified

    @pytest.mark.asyncio
    async def test_last_modified_received(self):
        cache = {}
        last_modified = "12345"
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["last-modified"] = last_modified
        gh = MockGitHubAPI(200, headers, b"42", cache=cache)
        data = await gh.getitem("/fake")
        url = "https://api.github.com/fake"
        assert url in cache
        assert cache[url] == (None, last_modified, 42, None)
        assert data == cache[url][2]

    @pytest.mark.asyncio
    async def test_hit(self):
        url = "https://api.github.com/fake"
        cache = {url: ("12345", "67890", 42, None)}
        gh = MockGitHubAPI(304, cache=cache)
        data = await gh.getitem(url)
        assert data == 42

    @pytest.mark.asyncio
    async def test_miss(self):
        url = "https://api.github.com/fake"
        cache = {url: ("12345", "67890", 42, None)}
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["etag"] = "09876"
        headers["last-modified"] = "54321"
        gh = MockGitHubAPI(200, headers, body=b"-13", cache=cache)
        data = await gh.getitem(url)
        assert data == -13
        assert cache[url] == ("09876", "54321", -13, None)

    @pytest.mark.asyncio
    async def test_ineligible(self):
        cache = {}
        gh = MockGitHubAPI(cache=cache)
        url = "https://api.github.com/fake"
        # Only way to force a GET request with a body.
        await gh._make_request("GET", url, {}, 42, "asdf")
        assert url not in cache
        await gh.post(url, data=42)
        assert url not in cache

    @pytest.mark.asyncio
    async def test_redirect_without_cache(self):
        cache = {}
        gh = MockGitHubAPI(304, cache=cache)
        with pytest.raises(RedirectionException):
            await gh.getitem("/fake")

    @pytest.mark.asyncio
    async def test_no_cache(self):
        headers = MockGitHubAPI.DEFAULT_HEADERS.copy()
        headers["etag"] = "09876"
        headers["last-modified"] = "54321"
        gh = MockGitHubAPI(headers=headers)
        await gh.getitem("/fake")  # No exceptions raised.


_SAMPLE_QUERY = """query {
    viewer {
        name
        repositories(last: 3) {
            nodes {
                name
            }
        }
    }
}
"""

_SAMPLE_QUERY_WITH_VARIABLES = """query ($number_of_repos: Int!) {
    viewer {
        name
        repositories(last: $number_of_repos) {
            nodes {
                name
            }
        }
    }
}
"""


class TestGraphQL:
    """Test gidgethub.abc.GitHubAPI.graphql()."""

    def gh_and_response(self, payload_filename):
        payload = (
            importlib_resources.files(graphql_samples) / payload_filename
        ).read_bytes()
        status_code_match = re.match(r"^.+-(\d+)\.json$", payload_filename)
        status_code = int(status_code_match.group(1))
        return (
            MockGitHubAPI(status_code, body=payload, oauth_token="oauth-token"),
            json.loads(payload.decode("utf-8")),
        )

    @pytest.mark.asyncio
    async def test_5XX_status_code(self):
        response_data = {"hello": "World"}
        gh = MockGitHubAPI(
            500,
            body=json.dumps(response_data).encode("utf-8"),
            oauth_token="oauth-token",
        )
        with pytest.raises(GitHubBroken) as exc:
            await gh.graphql("does not matter")
        assert exc.value.status_code == http.HTTPStatus(500)

    @pytest.mark.asyncio
    async def test_bad_credentials(self):
        gh, response_data = self.gh_and_response("bad-credentials-401.json")
        with pytest.raises(GraphQLAuthorizationFailure) as exc:
            await gh.graphql(_SAMPLE_QUERY)
        assert exc.value.response == response_data
        assert exc.value.status_code == http.HTTPStatus(401)

    @pytest.mark.asyncio
    async def test_4XX_status_code(self):
        """Test a 4XX response for which there is no pre-defined exception.

        The testing of the response to sending bad JSON is inconsequential.
        """
        gh, response_data = self.gh_and_response("malformed-json-400.json")
        with pytest.raises(BadGraphQLRequest) as exc:
            await gh.graphql(_SAMPLE_QUERY)
        assert exc.value.response == response_data
        assert exc.value.status_code == http.HTTPStatus(400)

    @pytest.mark.asyncio
    async def test_malformed_query(self):
        gh, response_data = self.gh_and_response("malformed-query-200.json")
        with pytest.raises(QueryError) as exc:
            await gh.graphql("this isn't right")
        assert exc.value.response == response_data

    @pytest.mark.asyncio
    async def test_missing_variable(self):
        gh, response_data = self.gh_and_response("missing-variable-in-request-200.json")
        with pytest.raises(QueryError) as exc:
            await gh.graphql(_SAMPLE_QUERY_WITH_VARIABLES)
        assert exc.value.response == response_data

    @pytest.mark.asyncio
    async def test_unexpected_200_response(self):
        """200 responses are expected to contain either an "errors" or "data" key."""
        response_data = {"hello": "World"}
        gh = MockGitHubAPI(
            200,
            body=json.dumps(response_data).encode("utf-8"),
            oauth_token="oauth-token",
        )
        with pytest.raises(GraphQLException) as exc:
            await gh.graphql("does not matter")
        assert exc.value.response == response_data

    @pytest.mark.asyncio
    async def test_query(self):
        gh, response_data = self.gh_and_response("success-200.json")
        result = await gh.graphql(_SAMPLE_QUERY)
        assert gh.method == "POST"
        assert gh.url == "https://api.github.com/graphql"
        assert gh.headers["content-type"] == "application/json; charset=utf-8"
        assert gh.headers["content-length"] == str(len(gh.body))
        expected_headers = sansio.create_headers(
            "test_abc",
            accept="application/json; charset=utf-8",
            oauth_token="oauth-token",
        )
        for key, value in expected_headers.items():
            assert gh.headers[key] == value
        body = json.loads(gh.body.decode("utf-8"))
        assert body == {"query": _SAMPLE_QUERY}
        assert result == response_data["data"]

    @pytest.mark.asyncio
    async def test_query_with_variables(self):
        gh, response_data = self.gh_and_response("success-200.json")
        result = await gh.graphql(_SAMPLE_QUERY_WITH_VARIABLES, number_of_repos=3)
        assert gh.method == "POST"
        assert gh.url == "https://api.github.com/graphql"
        assert gh.headers["content-type"] == "application/json; charset=utf-8"
        assert gh.headers["content-length"] == str(len(gh.body))
        expected_headers = sansio.create_headers(
            "test_abc",
            accept="application/json; charset=utf-8",
            oauth_token="oauth-token",
        )
        for key, value in expected_headers.items():
            assert gh.headers[key] == value
        body = json.loads(gh.body.decode("utf-8"))
        assert body == {
            "query": _SAMPLE_QUERY_WITH_VARIABLES,
            "variables": {"number_of_repos": 3},
        }
        assert result == response_data["data"]

    @pytest.mark.asyncio
    async def test_endpoint(self):
        gh, response_data = self.gh_and_response("success-200.json")
        base_url = "https://example.com/graphql"
        result = await gh.graphql(_SAMPLE_QUERY, endpoint=base_url)
        assert gh.method == "POST"
        assert gh.url == base_url
        assert gh.headers["content-type"] == "application/json; charset=utf-8"
        assert gh.headers["content-length"] == str(len(gh.body))
        expected_headers = sansio.create_headers(
            "test_abc",
            accept="application/json; charset=utf-8",
            oauth_token="oauth-token",
        )
        for key, value in expected_headers.items():
            assert gh.headers[key] == value
        body = json.loads(gh.body.decode("utf-8"))
        assert body == {"query": _SAMPLE_QUERY}
        assert result == response_data["data"]

    @pytest.mark.asyncio
    async def test_unexpected_status_code(self):
        response_data = {"hello": "World"}
        gh = MockGitHubAPI(
            300,
            body=json.dumps(response_data).encode("utf-8"),
            oauth_token="oauth-token",
        )
        with pytest.raises(GraphQLException) as exc:
            await gh.graphql("does not matter")
        assert exc.value.response == response_data

    @pytest.mark.asyncio
    async def test_response_content_type_parsing_gh121(self):
        gh, response_data = self.gh_and_response("success-200.json")
        # Test that a JSON content type still works if formatted without spaces.
        gh.response_headers["content-type"] = "application/json;charset=utf-8"
        # Should not fail.
        resp = await gh.graphql("does not matter")
        assert resp is not None
        # Spaces in the content-type header should not be a problem.
        gh.response_headers["content-type"] = "application/json; charset=utf-8"
        # Should not fail.
        resp = await gh.graphql("does not matter")
        assert resp is not None

    @pytest.mark.asyncio
    async def test_unknown_response_content_type_gh121(self):
        gh, response_data = self.gh_and_response("success-200.json")
        # A non-JSON response should raise an exception.
        gh.response_headers["content-type"] = "application/gidget;charset=utf-8"
        with pytest.raises(GraphQLResponseTypeError):
            await gh.graphql("does not matter")

    @pytest.mark.asyncio
    async def test_no_response_content_type_gh121(self):
        gh, response_data = self.gh_and_response("success-200.json")
        # An empty content type should raise an exception.
        gh.response_headers["content-type"] = ""
        with pytest.raises(GraphQLException):
            await gh.graphql("does not matter")
        del gh.response_headers["content-type"]
        with pytest.raises(GraphQLException):
            await gh.graphql("does not matter")

    @pytest.mark.asyncio
    async def test_no_response_data(self):
        # An empty response should raise an exception.
        gh = MockGitHubAPI(200, body=b"", oauth_token="oauth-token")
        with pytest.raises(GraphQLException):
            await gh.graphql("does not matter")



"""Provide an abstract base class for easier requests."""

import abc
import http
import json
from typing import Any, AsyncGenerator, Dict, Mapping, MutableMapping, Optional, Tuple
from typing import Optional as Opt

from uritemplate import variable

from . import (
    BadGraphQLRequest,
    GitHubBroken,
    GraphQLAuthorizationFailure,
    GraphQLException,
    HTTPException,
    QueryError,
    GraphQLResponseTypeError,
)
from . import sansio


# Value represents etag, last-modified, data, and next page.
CACHE_TYPE = MutableMapping[str, Tuple[Opt[str], Opt[str], Any, Opt[str]]]

JSON_CONTENT_TYPE = "application/json"
UTF_8_CHARSET = "utf-8"
JSON_UTF_8_CHARSET = f"{JSON_CONTENT_TYPE}; charset={UTF_8_CHARSET}"
ITERABLE_KEY = "items"


class GitHubAPI(abc.ABC):
    """Provide an idiomatic API for making calls to GitHub's API."""

    def __init__(
        self,
        requester: str,
        *,
        oauth_token: Opt[str] = None,
        cache: Opt[CACHE_TYPE] = None,
        base_url: str = sansio.DOMAIN,
    ) -> None:
        self.requester = requester
        self.oauth_token = oauth_token
        self._cache = cache
        self.rate_limit: Opt[sansio.RateLimit] = None
        self.base_url = base_url

    @abc.abstractmethod
    async def _request(
        self, method: str, url: str, headers: Mapping[str, str], body: bytes = b""
    ) -> Tuple[int, Mapping[str, str], bytes]:
        """Make an HTTP request."""

    @abc.abstractmethod
    async def sleep(self, seconds: float) -> None:
        """Sleep for the specified number of seconds."""

    async def _make_request(
        self,
        method: str,
        url: str,
        url_vars: Optional[variable.VariableValueDict],
        data: Any,
        accept: str,
        jwt: Opt[str] = None,
        oauth_token: Opt[str] = None,
        content_type: str = JSON_CONTENT_TYPE,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> Tuple[bytes, Opt[str], int]:
        """Construct and make an HTTP request."""
        if oauth_token is not None and jwt is not None:
            raise ValueError("Cannot pass both oauth_token and jwt.")
        filled_url = sansio.format_url(url, url_vars, base_url=self.base_url)
        if jwt is not None:
            request_headers = sansio.create_headers(
                self.requester, accept=accept, jwt=jwt
            )
        elif oauth_token is not None:
            request_headers = sansio.create_headers(
                self.requester, accept=accept, oauth_token=oauth_token
            )
        else:
            # fallback to using oauth_token
            request_headers = sansio.create_headers(
                self.requester, accept=accept, oauth_token=self.oauth_token
            )
        if extra_headers is not None:
            request_headers.update(extra_headers)
        cached = cacheable = False
        # Can't use None as a "no body" sentinel as it's a legitimate JSON type.
        if data == b"":
            body = b""
            request_headers["content-length"] = "0"
            if method == "GET" and self._cache is not None:
                cacheable = True
                try:
                    etag, last_modified, data, more = self._cache[filled_url]
                    cached = True
                except KeyError:
                    pass
                else:
                    if etag is not None:
                        request_headers["if-none-match"] = etag
                    if last_modified is not None:
                        request_headers["if-modified-since"] = last_modified
        else:
            if content_type != JSON_CONTENT_TYPE:
                # We don't know how to handle other content types, so just pass things along.
                request_headers["content-type"] = content_type
                body = data
            else:
                # Since JSON is so common, add some niceties.
                body = json.dumps(data).encode(UTF_8_CHARSET)
                request_headers["content-type"] = JSON_UTF_8_CHARSET
            request_headers["content-length"] = str(len(body))
        if self.rate_limit is not None:
            self.rate_limit.remaining -= 1
        response = await self._request(method, filled_url, request_headers, body)
        if not (response[0] == 304 and cached):
            data, self.rate_limit, more = sansio.decipher_response(*response)
            has_cache_details = "etag" in response[1] or "last-modified" in response[1]
            if self._cache is not None and cacheable and has_cache_details:
                etag = response[1].get("etag")
                last_modified = response[1].get("last-modified")
                self._cache[filled_url] = etag, last_modified, data, more
        return data, more, response[0]

    async def getitem(
        self,
        url: str,
        url_vars: Optional[variable.VariableValueDict] = {},
        *,
        accept: str = sansio.accept_format(),
        jwt: Opt[str] = None,
        oauth_token: Opt[str] = None,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        """Send a GET request for a single item to the specified endpoint."""

        data, _, _ = await self._make_request(
            "GET",
            url,
            url_vars,
            b"",
            accept,
            jwt=jwt,
            oauth_token=oauth_token,
            extra_headers=extra_headers,
        )
        return data

    async def getstatus(
        self,
        url: str,
        url_vars: Optional[variable.VariableValueDict] = {},
        *,
        accept: str = sansio.accept_format(),
        jwt: Opt[str] = None,
        oauth_token: Opt[str] = None,
    ) -> int:
        """Send a GET request for a single item to the specifie endpoint and return its status code."""

        try:
            _, _, status_code = await self._make_request(
                "GET", url, url_vars, b"", accept, jwt=jwt, oauth_token=oauth_token
            )
        except HTTPException as e:
            status_code = e.status_code

        return status_code

    async def getiter(
        self,
        url: str,
        url_vars: Optional[variable.VariableValueDict] = {},
        *,
        accept: str = sansio.accept_format(),
        jwt: Opt[str] = None,
        oauth_token: Opt[str] = None,
        extra_headers: Optional[Dict[str, str]] = None,
        iterable_key: Opt[str] = ITERABLE_KEY,
    ) -> AsyncGenerator[Any, None]:
        """Return an async iterable for all the items at a specified endpoint."""
        current_url: Opt[str] = url
        while current_url:
            data, current_url, _ = await self._make_request(
                "GET",
                current_url,
                url_vars,
                b"",
                accept,
                jwt=jwt,
                oauth_token=oauth_token,
                extra_headers=extra_headers,
            )

            if isinstance(data, dict) and iterable_key in data:
                data = data[iterable_key]
            for item in data:
                yield item

    async def post(
        self,
        url: str,
        url_vars: Optional[variable.VariableValueDict] = {},
        *,
        data: Any,
        accept: str = sansio.accept_format(),
        jwt: Opt[str] = None,
        oauth_token: Opt[str] = None,
        extra_headers: Optional[Dict[str, str]] = None,
        content_type: str = JSON_CONTENT_TYPE,
    ) -> Any:
        data, _, _ = await self._make_request(
            "POST",
            url,
            url_vars,
            data,
            accept,
            jwt=jwt,
            oauth_token=oauth_token,
            content_type=content_type,
            extra_headers=extra_headers,
        )
        return data

    async def patch(
        self,
        url: str,
        url_vars: Optional[variable.VariableValueDict] = {},
        *,
        data: Any,
        accept: str = sansio.accept_format(),
        jwt: Opt[str] = None,
        oauth_token: Opt[str] = None,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        data, _, _ = await self._make_request(
            "PATCH",
            url,
            url_vars,
            data,
            accept,
            jwt=jwt,
            oauth_token=oauth_token,
            extra_headers=extra_headers,
        )
        return data

    async def put(
        self,
        url: str,
        url_vars: Optional[variable.VariableValueDict] = {},
        *,
        data: Any = b"",
        accept: str = sansio.accept_format(),
        jwt: Opt[str] = None,
        oauth_token: Opt[str] = None,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        data, _, _ = await self._make_request(
            "PUT",
            url,
            url_vars,
            data,
            accept,
            jwt=jwt,
            oauth_token=oauth_token,
            extra_headers=extra_headers,
        )
        return data

    async def delete(
        self,
        url: str,
        url_vars: Optional[variable.VariableValueDict] = {},
        *,
        data: Any = b"",
        accept: str = sansio.accept_format(),
        jwt: Opt[str] = None,
        oauth_token: Opt[str] = None,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> None:
        await self._make_request(
            "DELETE",
            url,
            url_vars,
            data,
            accept,
            jwt=jwt,
            oauth_token=oauth_token,
            extra_headers=extra_headers,
        )

    async def graphql(
        self,
        query: str,
        *,
        endpoint: str = "https://api.github.com/graphql",
        **variables: Any,
    ) -> Any:
        """Query the GraphQL v4 API.

        The *endpoint* argument specifies the endpoint URL to use. The
        *variables* kwargs-style argument collects all variables for the query.
        """
        payload: Dict[str, Any] = {"query": query}
        if variables:
            payload["variables"] = variables
        request_data = json.dumps(payload).encode("utf-8")
        request_headers = sansio.create_headers(
            self.requester, accept=JSON_UTF_8_CHARSET, oauth_token=self.oauth_token
        )
        request_headers.update(
            {
                "content-type": JSON_UTF_8_CHARSET,
                "content-length": str(len(request_data)),
            }
        )
        status_code, response_headers, response_data = await self._request(
            "POST", endpoint, request_headers, request_data
        )

        if not response_data:
            raise GraphQLException("Response contained no data", response_data)

        # Decode content.
        resp_content_type = response_headers.get("content-type")
        type_, encoding = sansio._parse_content_type(resp_content_type)
        response_str = response_data.decode(encoding)
        if type_ == "application/json":
            response: Dict[str, Any] = json.loads(response_str)
        else:
            raise GraphQLResponseTypeError(resp_content_type, response_str)

        if status_code >= 500:
            raise GitHubBroken(http.HTTPStatus(status_code))
        elif status_code == 401:
            raise GraphQLAuthorizationFailure(response)
        elif status_code >= 400:
            # 400 corresponds to malformed JSON, but that should never receive
            # that as a response as json.dumps() should have raised its own
            # exception before we made the request.
            raise BadGraphQLRequest(http.HTTPStatus(status_code), response)
        elif status_code == 200:
            self.rate_limit = sansio.RateLimit.from_http(response_headers)
            if "errors" in response:
                raise QueryError(response)
            if "data" in response:
                return response["data"]
            else:
                raise GraphQLException(
                    f"Response did not contain 'errors' or 'data': {response}", response
                )
        else:
            raise GraphQLException(
                f"Unexpected HTTP response to GraphQL request: {status_code}", response
            )



import datetime

import aiohttp
import pytest

from gidgethub import aiohttp as gh_aiohttp
from gidgethub import sansio


@pytest.mark.asyncio
async def test_sleep():
    delay = 1
    start = datetime.datetime.now()
    async with aiohttp.ClientSession() as session:
        gh = gh_aiohttp.GitHubAPI(session, "gidgethub")
        await gh.sleep(delay)
    stop = datetime.datetime.now()
    assert (stop - start) > datetime.timedelta(seconds=delay)


@pytest.mark.asyncio
async def test__request():
    """Make sure that that abstract method is implemented properly."""
    request_headers = sansio.create_headers("gidgethub")
    async with aiohttp.ClientSession() as session:
        gh = gh_aiohttp.GitHubAPI(session, "gidgethub")
        aio_call = await gh._request(
            "GET", "https://api.github.com/rate_limit", request_headers
        )
    data, rate_limit, _ = sansio.decipher_response(*aio_call)
    assert "rate" in data


@pytest.mark.asyncio
async def test_get():
    """Integration test."""
    async with aiohttp.ClientSession() as session:
        gh = gh_aiohttp.GitHubAPI(session, "gidgethub")
        data = await gh.getitem("/rate_limit")
    assert "rate" in data
